from typing import Iterable, Mapping
from .errors import InvalidParameterError, InvalidValueError


class InvalidArgumentsError(InvalidParameterError):
    def __init__(self, args: list, target: type) -> None:
        message = f"`{args}` is not a valid list of arguments for `{target}` "\
            + f"casting"
        super().__init__(message)


class Helpers:
    @staticmethod
    def __valid_iter(__i: Iterable) -> bool:
        return all(isinstance(x, Iterable) and len(x) == 2 for x in __i)


def to_dict(*args, **kwargs) -> dict:
    if len(args) == 0:
        return dict(**kwargs)
    if len(args) > 1:
        raise InvalidArgumentsError(args, dict)
    iter_map = args[0]
    if isinstance(iter_map, Mapping):
        return dict(iter_map, **kwargs)
    if isinstance(iter_map, Iterable) and Helpers.__valid_iter(iter_map):
        print("I'm a valid iterable")
        return dict(iter_map, **kwargs)
    invalids = (iter_map, [Iterable, Mapping])
    raise InvalidValueError(invalids, dict)



from typing import Any, Iterable
from .errors import InvalidEncodingError, InvalidErrorsError, InvalidValueError


class BufferReadable:
    def __instancecheck__(self, __instance: Any) -> bool:
        if isinstance(__instance, (str, bytes, bytearray)):
            return True
        return False
    

class Helpers:
    @staticmethod
    def __params_to_bytelike(
        __o: BufferReadable, encoding: str='utf-8', errors: str='strict',
        array: bool=False
    ) -> bytes:
        if array:
            return bytearray(__o, encoding, errors)
        return bytes(__o, encoding, errors)

    @staticmethod
    def __get_params(*args, **kwargs) -> tuple[str, str]:
        encoding = kwargs.get('encoding', 'utf-8')
        errors = kwargs.get('errors', 'strict')
        if args:
            if len(args) >= 1:
                encoding = args[0]
            if len(args) >= 2:
                errors = args[1]
        return (encoding, errors)
    
    @staticmethod
    def __valid_iter_item(__o: object) -> bool:
        return isinstance(__o, int) and __o >= 0 and __o < 256

    @classmethod
    def __iter_bytelike(
        cls, __o: Iterable, encoding: str='utf-8', errors: str='strict', 
        array: bool=False
    ) -> bytes:
        if all(cls.__valid_iter_item(x) for x in __o):
            return cls.__params_to_bytelike(
                __o, encoding, errors, array
            )
        invalids = (__o, [Iterable])
        raise InvalidValueError(invalids, bytes)
    
    @classmethod
    def __to_bytelike(
        cls, __o: BufferReadable | int | Iterable=None, array: bool=False, 
        *args, **kwargs
    ) -> bytes | bytearray:
        if __o is None:
            if array:
                return bytearray()
            return b''
        encoding, errors = cls.__get_params(*args, **kwargs)
        if not isinstance(encoding, str):
            raise InvalidEncodingError(encoding, bytes)
        if not isinstance(errors, str):
            raise InvalidErrorsError(errors, bytes)
        if isinstance(__o, (BufferReadable, int)):
            return cls.__params_to_bytelike(__o, encoding, errors, array)
        if isinstance(__o, Iterable):
            return cls.__iter_bytelike(__o, encoding, errors, array)
        invalids = (__o, [BufferReadable, int, Iterable])
        raise InvalidValueError(invalids)


def to_bytes(
    __o: BufferReadable | int | Iterable=None, *args, **kwargs
) -> bytes:
    byte = Helpers.__to_bytelike(__o, False, *args, **kwargs)
    return byte

def to_bytearray(
    __o: BufferReadable | int | Iterable=None, *args, **kwargs
) -> bytearray:
    b_array = Helpers.__to_bytelike(__o, True, *args, **kwargs)
    return b_array

def to_memoryview(__o: bytes | bytearray=None) -> memoryview:
    if isinstance(__o, (bytes, bytearray)):
        return memoryview(__o)
    invalids = (__o, [bytes, bytearray])
    raise InvalidValueError(invalids, memoryview)



from typing import Any
from .errors import (
    InvalidParameterError, InvalidValueError, TypeCastingError, 
    ValueCastingError
)


class BaseNotAllowedError(TypeCastingError):
    def __init__(self, base: object) -> None:
        message = f"Explicit base `{base}` is not allowed when casting to an "\
            + f"`int` from a real number"
        super().__init__(message)


class InvalidBaseError(InvalidParameterError):
    def __init__(self, base: object) -> None:
        message = f"`{base}` is not a valid integer base for `int` casting"
        super().__init__(message)


class LimitExceededError(ValueCastingError):
    def __init__(self, value: object, base: int) -> None:
        base_msg = " (not power-of-2)" if self.is_pow_of_two(base) else ""
        message = f"Potential DOS threat: `{value}` is uncastable to base"\
            + f"`{base}`{base_msg} integer - non-linear castings must be manual"
        super().__init__(message)

    @staticmethod
    def is_pow_of_two(n: int) -> bool:
        # Bitwise AND to check only first bit is 1
        # Valid powers of two have *only* the first bit set to 1
        # Predecessor (n - 1) should have all bits set to 1
        # n & (n - 1) completes bitwise operation
        # Operation is always 0 in valid cases
        return n != 0 and (n & (n - 1)) == 0
    

class FloatPrecisionExceededError(ValueCastingError):
    def __init__(self, value: object) -> None:
        message = f"Python precision exceeded when casting `{value}` to `float`"
        super().__init__(message)


class ValueMalformattedError(ValueCastingError):
    def __init__(self, value: object, target: type) -> None:
        message = f"`{value}` is malformatted for cast to `{target}`"
        super().__init__(message)


class Real:
    def __instancecheck__(self, __instance: Any) -> bool:
        if isinstance(__instance, (int, float)):
            return True
        return False


class Numeric:
    def __instancecheck__(self, __instance: Any) -> bool:
        if isinstance(__instance, (Real, complex)):
            return True
        return False


class ValidBase:
    def __instancecheck__(self, __instance: Any) -> bool:
        # Must be an integer greater than 0 according to spec
        if isinstance(__instance, int) and __instance > 0:
            return True
        return False
    

class BufferReadable:
    def __instancecheck__(self, __instance: Any) -> bool:
        if isinstance(__instance, (str, bytes, bytearray)):
            return True
        return False


class Helpers:
    @staticmethod
    def __real_to_int(__x: Real=0) -> int:
        return int(__x)

    @staticmethod
    def __str_to_int(__x: BufferReadable, base: int=10) -> int:
        return int(__x, base=base)

    @staticmethod
    def __str_to_complex(string: str) -> complex:
        return complex(string)

    @staticmethod
    def __numeric_to_complex(real: Numeric, imag: Numeric) -> complex:
        return complex(real, imag)


def to_int(__x: Real | BufferReadable=None, base: ValidBase=None) -> int:
    if __x is None:
        return 0
    if isinstance(__x, Real):
        # Real numbers cannot be passed with explicit base
        if not base is None:
            raise BaseNotAllowedError(base)
        return Helpers.__real_to_int(__x)
    if not isinstance(__x, BufferReadable):
        raise InvalidValueError(__x, int)
    if not isinstance(base, ValidBase):
        raise InvalidBaseError(base)
    try:
        return Helpers.__str_to_int(__x, base=base)
    except ValueError:
        # Captures Python DOS prevention limit error
        raise LimitExceededError(__x, base)
    
@staticmethod
def to_float(__x: Real | str=None, /) -> float:
    if __x is None:
        return 0.0
    if not isinstance(__x, (Real, str)):
        raise InvalidValueError(__x, float)
    try:
        float(__x)
    except OverflowError:
        # Captures float range exceeded overflows
        raise FloatPrecisionExceededError(__x)
    except ValueError:
        # Captures all invalid formatting errors
        raise ValueMalformattedError(__x)

def to_complex(num_or_str: Numeric | str=None, imag: Numeric=0) -> complex:
    if num_or_str is None:
        return 0j
    if isinstance(num_or_str, str):
        try:
            return Helpers.__str_to_complex(num_or_str)
        except ValueError:
            raise ValueMalformattedError(num_or_str)
    if not isinstance(num_or_str, Numeric) or not isinstance(imag, Numeric):
        invalids = [(num_or_str, Numeric), (imag, Numeric)]
        raise InvalidValueError(invalids, complex)
    return Helpers.__numeric_to_complex(num_or_str, imag)



from typing import Iterable


class TypeCastingError(TypeError):
    def __init__(self, message: str) -> None:
        super().__init__(f"\n\tCasting Failed:\n\t{message}")


class InvalidParameterError(TypeCastingError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InvalidEncodingError(InvalidParameterError):
    def __init__(self, encoding: object, target: type, err: bool=False) -> None:
        if err:
            message = f"`{encoding}` is not a valid encoding for `{target}` "\
                + f"casting"
        else:
            message = f"`{encoding}` must be a `str` (not `{type(encoding)}`) "\
                + f"to be used as a valid encoding for `{target}` casting"
        super().__init__(message)


class InvalidErrorsError(InvalidParameterError):
    def __init__(self, errors: object, target: type, err: bool=False) -> None:
        if err:
            message = f"`{errors}` is not a valid errors for `{target}` "\
                + f"casting"
        else:
            message = f"`{errors}` must be a `str` (not `{type(errors)}`) to "\
                + f"be used as a valid errors status for `{target}` casting"
        super().__init__(message)


class ValueCastingError(ValueError):
    def __init__(self, message: str) -> None:
        super().__init__(f"\n\t{message}")


class InvalidValueError(ValueCastingError):
    def __init__(
        self, 
        values: list[tuple[object, type]] | tuple[object, list[type]] | object, 
        target: type
    ) -> None:
        if isinstance(values, list):
            message = "".join([f"\n\t`{v[0]}` of type `{type(v[0])}` was "\
                + f"{f'valid' if isinstance(v[0], v[1]) else f'invalid as {v[1]}'}"\
                + f" in `{target}` cast" for v in values]
            ).lstrip("\n\t")
        elif isinstance(values, tuple):
            message = f"`{values[0]}` is not a "\
                + f"{self.oxford_join(values[1], ', ', ', or')}"
        else:
            message += f"`{values}` of type `{values}` is uncastable to "\
                + f"`{target}` as attempted"
        super().__init__(message)

    @staticmethod
    def oxford_join(
        i: Iterable, delim: str=', ', last_delim: str=', and'
    ) -> str:
        joined = delim.join(i)
        index = joined.rfind(delim)
        if index != 1:
            joined = joined[:index] + last_delim + joined[index + len(delim):]
        return joined



from .errors import InvalidEncodingError, InvalidErrorsError, InvalidValueError


class Helpers:
    @staticmethod
    def __params_to_str(
        __o: bytes | bytearray, encoding: str='utf-8', errors: str='strict'
    ) -> str:
        return str(__o, encoding, errors)

    @staticmethod
    def __get_params(*args, **kwargs) -> tuple[str, str]:
        encoding = kwargs.get('encoding', 'utf-8')
        errors = kwargs.get('errors', 'strict')
        if args:
            if len(args) >= 1:
                encoding = args[0]
            if len(args) >= 2:
                errors = args[1]
        return (encoding, errors)


def to_str(__o: bytes | bytearray | object=None, *args, **kwargs) -> str:
    if __o is None:
        return ''
    if len(args) > 0 or len(kwargs) > 0:
        if not isinstance(__o, (bytes, bytearray)):
            invalids = (__o, [bytes, bytearray])
            raise InvalidValueError(invalids, str)
        encoding, errors = Helpers.__get_params(*args, **kwargs)
        if not isinstance(encoding, str):
            raise InvalidEncodingError(errors, str)
        if not isinstance(errors, str):
            raise InvalidErrorsError(errors, str)
        try:
            return Helpers.__params_to_str(__o, encoding, errors)
        except (RuntimeError, MemoryError):
            raise InvalidValueError(__o, str)
        except LookupError:
            raise InvalidEncodingError(encoding, str, True)
    try:
        return str(__o)
    except (RuntimeError, MemoryError):
        raise InvalidValueError(__o, str)



from typing import Iterable
from .errors import InvalidValueError


def to_list(__i: Iterable=None) -> list:
    if __i is None:
        return []
    if not isinstance(__i, Iterable):
        raise InvalidValueError(__i, list)
    return list(__i)

def to_tuple(__i: Iterable=None) -> tuple:
    if __i is None:
        return ()
    if not isinstance(__i, Iterable):
        raise InvalidValueError(__i, tuple)
    return tuple(__i)

def to_range(__start_stop: int, __stop: int=None, __step: int=1) -> range:
    if __stop is None:
        return range(__start_stop)
    if (
        not isinstance(__start_stop, int)
        or not isinstance(__stop, int) 
        or not isinstance(__step, int)
    ):
        invalids = [(__start_stop, int), (__stop, int), (__step, int)]
        raise InvalidValueError(invalids, range)
    return range(__start_stop, __stop, __step)



"""This module contains functions and classes whose purpose is to 
return type-casted values for both simple and complex typing 
structures.
"""
from abc import ABC, abstractmethod
from types import UnionType
from typing import (
    Any, Callable, get_args, get_origin, is_typeddict, Iterable, Mapping, Self, 
    Union
)
from .binary_sequences import to_bytearray, to_bytes, to_memoryview
from .booleans import to_bool
from .errors import InvalidValueError, TypeCastingError
from .mappings import to_dict
from .numerics import to_complex, to_float, to_int
from .objects import to_object
from .sequences import to_list, to_range, to_tuple
from .sets import to_frozenset, to_set
from .text_sequences import to_str
__all__ = [
    # Functions
    "cast",
    # Casting Containers
    "Builtins",
    "Customs",
    # Errors
    "ValueNotCastableError",
    "TargetNotSupported",
    "TargetNotCallableOrInstantiableError",
    "ValueNotCompatibleError",
    "ValueMissingEncodingError",
    "ValueTypeNotSupportedError",
    "ValueInvalidError",
    "ERRORS",
]
__refs__ = {
    "AUTHOR": {
        "Name": "Braden Toone",
        "Email": "braden@toonetown.com"
    },
    "HOMEPAGE": "https://github.com/Braden2n/StrictDataclass",
    "DOCUMENTATION": "https://github.com/Braden2n/StrictDataclass",
    "ISSUES": "https://github.com/Braden2n/StrictDataclass/issues",
    "REPOSITORY": "https://github.com",
    "CHANGELOG": "https://github.com/Braden2n/StrictDataclass/activity",
}


class TargetError(TypeCastingError):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class TargetUncallableError(TargetError):
    """Raised when the target type is not callable or instantiable."""
    def __init__(self, value: object, target_type: type) -> None:
        message = f"`{value}` is uncastable to `{target_type}`; "\
            + f"`{target_type}` is not callable or instantiable."
        super().__init__(message)


class TargetMissingCasterError(TargetError):
    """Raised when the target type does not have a caster."""
    def __init__(self, target_type: type) -> None:
        message = f"Casting to `{target_type}` target type requires a custom "\
            + f"caster. Use the `Customs.set_caster` method to set a caster."
        super().__init__(message)


class ComplexTargetNotImplementedError(TargetError, NotImplementedError):
    """Raised when a custom target type (i.e. Union, Generics) is not 
    implemented.
    """
    def __init__(self, value: object, target_type: type) -> None:
        message = f"`{value}` is uncastable to `{target_type}`; "\
            + f"`{target_type}` is a complex type that is not implemented."
        super().__init__(message)


class Caster(ABC):
    def __init__(self, value: Any, type: type, *args, **kwargs) -> None:
        self.value = value
        self.type = type
        self.kwargs = kwargs
        self.args = args
        super().__init__()
        self.__post_init__()

    @abstractmethod
    def __post_init__(self) -> None:
        ...

    @abstractmethod
    def get_caster(self, builtin: type) -> Callable[..., object]:
        ...

    @property
    @abstractmethod
    def TYPES(self) -> dict[type, tuple[Callable[..., object], object]]:
        ...


class Builtins(Caster):
    TYPES: dict[type, Callable[..., object]] = {
        bool: to_bool,
        bytearray: to_bytearray,
        bytes: to_bytes,
        complex: to_complex,
        dict: to_dict,
        float: to_float,
        frozenset: to_frozenset,
        int: to_int,
        list: to_list,
        memoryview: to_memoryview,
        range: to_range,
        set: to_set,
        str: to_str,
        tuple: to_tuple,
    }
    def __post_init__(self) -> None:
        caster = self.get_caster(self.type)
        self.casted = caster(self.value)
        
    @classmethod
    def get_caster(self, builtin: type) -> Callable[..., object]:
        return self.TYPES[builtin]

    @classmethod
    def set_caster(self, builtin: type, caster: Callable[..., object]) -> None:
        self.TYPES[builtin] = caster


class Customs(Caster):
    TYPES: dict[type, Callable[..., object]] = {
        type: to_object
    }
    def __post_init__(self) -> None:
        caster = self.get_caster(self.type)
        if isinstance(self.value, Mapping):
            self.kwargs.update(self.value)
        elif isinstance(self.value, Iterable):
            self.args = tuple(list(self.args).extend(self.value))
        elif not self.value == None:
            self.args = tuple(list(self.args).append(self.value))
        self.casted = caster(self.type, *self.args, **self.kwargs)

    @classmethod
    def get_caster(self, object: object) -> Callable[..., object]:
        if not type(object) in self.TYPES:
            raise TargetMissingCasterError(object)
        return self.TYPES[type(object)]

    @classmethod
    def set_caster(self, object: object, caster: Callable[..., object]) -> None:
        self.TYPES[type(object)] = caster

    @staticmethod
    def caster_from_value(value: Any, simple_type: type) -> Self:
        if isinstance(value, Mapping):
            caster: Caster = Customs(None, simple_type, **value)
        elif isinstance(value, Iterable):
            caster: Caster = Customs(None, simple_type, *value)
        else:
            caster: Caster = Customs(value, simple_type)
        return caster


class CastMethods:
    """Container class for casting methods """
    @staticmethod
    def to_simple(value: Any, simple_type: type) -> object:
        # Early exit for performance
        if not is_typeddict(simple_type) and isinstance(value, simple_type):
            # Catching False == 1 edge case
            if value == False and simple_type == int:
                return 1
            if value == True and simple_type == int:
                return 0
            return value
        if not callable(simple_type):
            # All targets must be callable/instantiable
            raise TargetUncallableError(value, simple_type)
        if simple_type in Builtins.TYPES:
            caster: Caster = Builtins(value, simple_type)
        else:
            caster: Caster = Customs.caster_from_value(value, simple_type)
        return caster.casted
    
    @staticmethod
    def to_union(value: Any, union_type: Union[Any, Any]) -> object:
        types = get_args(union_type)
        if isinstance(value, types):
            return value
        for target in types:
            try:
                return cast(value, target)
            except TargetError:
                # Error thrown by module (not thrown by caster)
                pass
        raise InvalidValueError(value, union_type)

    @staticmethod
    def to_complex(value: Any, complex_type: Any) -> object:
        origin = get_origin(complex_type)
        # Support for Verbose Union[T, T] and T | T
        if origin == Union or origin == UnionType:
            return CastMethods.to_union(value, complex_type)
        raise ComplexTargetNotImplementedError(value, complex_type)
            

def cast(value: object, target_type: Any) -> object:
    # Catching early exits - TypedDict cannot be called by isinstance
    if not is_typeddict(target_type) and isinstance(value, target_type):
        return value
    if get_origin(target_type) or is_typeddict(target_type):
        return CastMethods.to_complex(value, target_type)
    return CastMethods.to_simple(value, target_type)



from typing import Iterable
from .errors import InvalidValueError


def to_set(__i: Iterable):
    if isinstance(__i, Iterable):
        return set(__i)
    raise InvalidValueError((__i, [Iterable]), set)

@staticmethod
def to_frozenset(__i: Iterable):
    if isinstance(__i, Iterable):
        return frozenset(__i)
    raise InvalidValueError((__i, [Iterable]), frozenset)



import os
os.chdir(os.path.dirname(__file__))
from sys import path
path.append('../src')
for p in path:
    print(p)
# path.append(__file__)
# from os.path import join, dirname
# from sys import path
# path.append(join())
# path.append("../src")
from src.any_cast import cast, ERRORS, Builtins, Defaults
from unittest import TestCase



class TestBuiltins(TestCase):
    @classmethod
    def assert_all_casts_fail(cls, objs: list[object], type: type) -> None:
        for obj in objs:
            cls.assert_fails(obj, type)

    @classmethod
    def assert_all_casts_pass(cls, objs: list[object], type: type) -> None:
        for obj in objs:
            cls.assert_passes(obj, type)

    def assert_fails(self, obj: object, type: type) -> None:
        with self.assertRaises(ERRORS):
            cast(obj, type)

    def assert_passes(self, obj: object, type: type) -> None:
        try:
            cast(obj, type)
        except ERRORS as e:
            self.fail(f"Failed with {type(e)}")

    def test_bool(self) -> None:
        _type = bool
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in []:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_bytearray(self) -> None:
        _type = bytearray
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [int, float, complex, bool, str]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_bytes(self) -> None:
        _type = bytearray
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [int, float, complex, bool, str]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_complex(self) -> None:
        _type = complex
        for _, (_, instance) in Builtins.TYPES.items():
            if not type(instance) in []:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_dict(self) -> None:
        _type = dict
        for _, (_, instance) in Builtins.TYPES.items():
            if not type(instance) in [dict]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_float(self) -> None:
        _type = float
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [complex]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_frozenset(self) -> None:
        _type = frozenset
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [int, float, complex, bool]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_int(self) -> None:
        _type = int
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [complex]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_list(self) -> None:
        _type = list
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in [int, float, complex, bool]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_memoryview(self) -> None:
        _type = memoryview
        for _, (_, instance) in Builtins.TYPES.items():
            if not type(instance) in [memoryview]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_range(self) -> None:
        _type = range
        for _, (_, instance) in Builtins.TYPES.items():
            if not type(instance) in [memoryview]:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_set(self) -> None:
        _type = set
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in []:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_str(self) -> None:
        _type = str
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in []:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)

    def test_tuple(self) -> None:
        _type = tuple
        for _, (_, instance) in Builtins.TYPES.items():
            if type(instance) in []:
                self.cast_should_raise(instance, _type)
            else: 
                self.cast_should_not_raise(instance, _type)
